;; buffer.asm - standard I/O buffer routines

;; Notes:
;;	So far we have character buffers.  How about word & dword buffers?
;;
;;	Error handling - what do we do for buffer overflow/underflow?


%include "memory.inc"
%include "odb.mac"
global	new_buffer, read_buf, write_buf


;; Buffer format:
;;	dd	head
;;	dd	tail
;;	dd	bottom		;last byte of buffer area
;;	dd	top		;first byte
;;	<buffer area>
%define head	0
%define tail	4
%define top	8
%define bottom	12

%define buf_hdr_size	16


	SECTION .text
init:	ret


new_buffer:
;;
;; Allocate & initialize a buffer
;;
;; Input:
%define .size	ebp+2*4	; Size of buffer
;;
;; Output:
%define	.addr	ebp+2*4	; Address (0 = failed)
;;
	push	ebp
	mov	ebp, esp
	pusha

	mov	ecx, [.size]
	mov	ebx, [DATA_end]		;address of header
	mov	eax, buf_hdr_size
	add	eax, ebx		;address of buffer area
	mov	[ebx+head], eax		;initialize head
	mov	[ebx+tail], eax		;initialize tail
	mov	[ebx+top], eax		;initialize top
	add	eax, ecx
	dec	eax
	mov	[ebx+bottom], eax	;initialize bottom
	add	ecx, buf_hdr_size
	add	[DATA_end], ecx		;allocate memory
	mov	[.addr], ebx		;return the address

	popa
	leave
	ret


write_buf:
;;
;; Write a character to a buffer
;;
;; Input:
;;	AL = character
;;	EBX = buffer (address of the header)
;;
	push	edi
	mov	edi, [ebx+head]	;head
	stosb
	cmp	edi, [ebx+bottom]	;at bottom?
	jbe	.1
	mov	edi, [ebx+top]	;wrap to top
.1	cmp	edi, [ebx+tail]	;buffer full? (if head=tail)
	je	.full
	mov	[ebx+head], edi	;save new head
.full	pop	edi
	ret


read_buf:
;;
;; Read a character from a buffer
;;
;; Input:
;;	EBX = buffer
;;
;; Output:
;;	AL = character
;;
	push	esi
	mov	esi, [ebx+tail]	;tail
	cmp	esi, [ebx+head]	;compare to head? If equal, buffer is empty.
	je	.empty
	lodsb	; al=[esi]; esi++;
	cmp	esi, [ebx+bottom]	;past bottom?
	jbe	.1
	mov	esi, [ebx+top]	;wrap around to top
.1	mov	[ebx+tail], esi	;save new tail
.end	pop	esi
	ret
.empty	xor	al, al		;For now, return 0 on error (not a great idea :) 
	jmp short .end		;Could return with CF set..
