; pmode.asm
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;

%include "memory.inc"
%include "boot.mac"
%include "macros.mac"
%include "forth.mac"

BITS 32

pmode:
	mov	ax, flat_data
	mov	ds, ax
	mov	es, ax
	mov	fs, ax
	mov	gs, ax
	mov	ss, ax
	mov	esp, RETURN_STACK
	mov	ebp, FORTH_STACK

	call	init_paging
	mov	esp, K_CODE

	mov	al, 11111011b	;disable all IRQs except cascade (IRQ 2)
	out	0x21, al
	mov	al, 11111111b
	out	0xA1, al
	sti
	call	init_screen 
	sprint "80x86 in 32-bit mode",10
	call	init_ints
	call	init_kbd
	call	init_floppy
	call	init_ide
        call    init_com
	call	init_dict
	call	init_forth
	call	BOOTSTRAP
	call	forth
;int 6
	sprint "hanging...",10
	jmp short $	;We should never get here, but stop just in case.

;------------------------------- Paging ----------------------------------

init_paging:
	;; Create page directory
	mov	edi, PAGE_TBL
	mov	eax, PAGE_TBL+0x1000+7	; 1st 4M linear->physical
	stosd
	mov	ecx, 0x400-2		; Unmapped
	xor	eax, eax
	rep	stosd
	mov	eax, PAGE_TBL+0x2000+7	; Last 4M for kernel
	stosd

	;; Create page table 0. Map the first 4M to physical RAM.
	mov	eax, 7		;Page table entry. P-W-U
	mov	ebx, 0x1000	;increment
	mov	ecx, 0x400	;count = 1k entries
				;edi = PAGE_TBL+0x1000
.loop	stosd
	add	eax, ebx
	loop	.loop

	;; Create page table FFF. Map last 4M for kernel modules, etc.
	xor	eax, eax
	mov	ecx, 0x400
	rep	stosd		;First, clear it
	;; Then map a few areas
%define map(l,p)  mov dword [PAGE_TBL+0x2000+(((l)>>10) & 0xFFC)], p+7
	map	(K_CODE,	0x100000)	;module code
	map	(K_CODE-0x1000,	0x101000)	;stack
	map	(K_DATA,	0x102000)	;module data

	mov	eax, PAGE_TBL	;Load PDBR (page dir. base reg.)
	mov	cr3, eax

	mov	eax, cr0
	or	eax, 0x80000000	;Set PG bit
	mov	cr0, eax
;	jmp short $+2		;always JMP after setting PG
		;(only required if the current page gets remapped?)
	ret




;-------------------------------------------------------------------------



%if 0
load_modules:
;
; TODO: Finish error handling. Stop at end of modules ("bad RDOFF2 header"
; error) but try to continue after some other error.
;
	push dword modules_start		;object
.loop
	;(Next object is already on stack, after the loop)
	push dword [kern_hdr]	;parent = 'kern'
	push dword 0				;flags
	call	load
	pop	eax		;Result code
	or	eax, eax
	jnz	.end		;Error - stop
	pop	eax		;.text address
phex eax
	pusha
	call	eax
	popa
dcr
	jmp short .loop
.end
	add	esp, 2*4
	ret
%endif



reboot:
;
; Currently using Clementine's method. Works great.
;
; Other methods:
;	Write a 0 to bit 0 of port 64h (8042 keyboard controller)
;
	cli
	mov	[0x472], word 0x1234	;BIOS flag: warm boot (no RAM test)

	mov	al, 0x8D		;Disable NMI & write to CMOS
	out	0x70, al
	in	al, 0x71

	mov	esi, 0			;Reload IDTR with a 0 limit
	mov	word [esi], 0		;Any interrupts will cause a
	lidt	[esi]			;triple fault (reboot)

	int	3			;Bam!
	jmp short $			;Hang if that fails



patch_int:
;
; Usage:
;	push	<int #>
;	push	<ISR address>
;	call	patch_int
;	add	esp, 8
;
	push	ebp
	mov	ebp, esp
	push	eax
	push	ebx

	mov	eax, [ebp+12]		; EAX = Interrupt number
	mov	ebx, [ebp+8]		; EBX = New ISR's address

	shl	eax, 3			; EAX *= 8
	add	eax, IDT_ADDRESS
	mov	[eax], bx		; Low-order word of offset
	shr	ebx, 16
	mov	[eax+6], bx		; High-order word

	pop	ebx
	pop	eax
	leave
	ret


%include "odb.mac"	;; temporary, just for flags..
%include "uvars.inc"
%include "dict.inc"
%include "forth.inc"
%include "ints.inc"
%include "display.inc"
%include "kbd.inc"
%include "ide.inc"
%include "graph.inc"
%include "serial.asm"
;%include "xcom.asm"
%include "floppy.asm"
